// lib.rs
//
// 此部分代码主要负责Rust-C两侧的数据内存结构转换，提供了C侧的函数接口。注意命名规范：
// 在C侧使用时，凡是函数名带有 '_cwallet'的，调用过之后都必须用'free_cwallet'释放内存，
// 否则导致内存泄漏
//

use qr::qr_code::QR;

use secp256k1::PublicKey;
use secp256k1::SecretKey;
use std::str::FromStr;
use utils::crypto_utils::{hash_password, scrypt_hash, verify_password, zdecrypt, zencrypt};
use utils::str_utils::{base64_to_hex, hex_to_base64};
use wasm_bindgen::prelude::*;

mod qr;
mod wallet;
use wallet_impl::Wallet;

mod utils;

use crate::wallet::*;

#[wasm_bindgen]
extern "C" {
    // Use `js_namespace` here to bind `console.log(..)` instead of just
    // `log(..)`
    #[wasm_bindgen(js_namespace = console)]
    fn log(s: &str);

    // The `console.log` is quite polymorphic, so we can bind it with multiple
    // signatures. Note that we need to use `js_name` to ensure we always call
    // `log` in JS.
    #[wasm_bindgen(js_namespace = console, js_name = log)]
    fn log_u32(a: u32);

    // Multiple arguments too!
    #[wasm_bindgen(js_namespace = console, js_name = log)]
    fn log_many(a: &str, b: &str);
}

macro_rules! console_log {
    // Note that this is using the `log` function imported above during
    // `bare_bones`
    ($($t:tt)*) => (log(&format_args!($($t)*).to_string()))
}

#[wasm_bindgen]
pub fn get_public_key(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    rwallet.get_public_key().to_string()
}

#[wasm_bindgen]
pub fn generate_sec_key(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    rwallet.generate_sec_key()
}

#[wasm_bindgen]
pub fn sign(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
    msg: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    let signature = rwallet.sign(msg);
    let r = match signature {
        Ok(v) => v,
        Err(err) => panic!("Problem sign: {:?}", err),
    };
    r
}

#[wasm_bindgen]
pub fn sign_for_tran(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
    msg: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };

    let signature = rwallet.sign_for_tran(msg);
    let (r, recid) = match signature {
        Ok((v, _recid)) => (v, _recid),
        Err(err) => panic!("Problem sign: {:?}", err),
    };
    let result = format!("{}|{}", r, recid);
    result
}

#[wasm_bindgen]
pub fn keccak256_hash(content: String) -> String {
    let msg_hex = utils::crypto_utils::keccak256_hash(&content);
    msg_hex
}

#[wasm_bindgen]
pub fn generate_scrypt_hash(
    pass: String,
    salt: String,
    n: f32,
    r: u32,
    p: u32,
    size: u32,
) -> String {
    let nr = n.log2().ceil() as u8;
    scrypt_hash(&pass, &salt, nr, r, p, size)
}

#[wasm_bindgen]
pub fn wget_address(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    // console_log!("wallet: {:?}", rwallet);
    let address = rwallet.get_address();
    let address_str = format!("{:?}", address);
    address_str
}

#[wasm_bindgen]
pub fn wencrypt(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
    msg: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    let pk = rwallet.get_public_key();
    let r = match zencrypt(pk, &msg) {
        Ok(v) => v,
        Err(err) => panic!("Problem encrypt: {:?}", err),
    };
    r
}
#[wasm_bindgen]
pub fn wdecrypt(
    id: String,
    openid: String,
    master_key: String,
    salt: String,
    pass: String,
    msg: String,
) -> String {
    let second_key = wallet_impl::generate_client_key(&pass, &openid, &salt);
    let msg_str = format!("{}{}{}", id, openid, salt);
    let msg_key = keccak256_hash(msg_str);
    let rwallet = Wallet {
        msg_key,
        master_key,
        second_key,
    };
    let pk = rwallet.get_secret_key();
    let r = match zdecrypt(pk, &msg) {
        Ok(v) => v,
        Err(err) => panic!("Problem encrypt: {:?}", err),
    };
    r
}

#[wasm_bindgen]
pub fn encrypt_content(content: String, key: String) -> String {
    let pass = key + "0x741482aE1480E552735E44Ff3A733448AcBbeD8d";
    let pass = utils::crypto_utils::keccak256_hash(&pass);
    let msg_hex = utils::crypto_utils::aes_encrypt(&content, &pass);
    msg_hex
}
#[wasm_bindgen]
pub fn decrypt_content(content: String, key: String) -> String {
    let pass = key + "0x741482aE1480E552735E44Ff3A733448AcBbeD8d";
    let pass = utils::crypto_utils::keccak256_hash(&pass);
    let msg_hex = utils::crypto_utils::aes_decrypt(&content, &pass);
    msg_hex
}
#[wasm_bindgen]
pub fn aes_encrypt(content: String, key: String) -> String {
    let pass = utils::crypto_utils::keccak256_hash(&key);
    let msg_hex = utils::crypto_utils::aes_encrypt(&content, &pass);
    msg_hex
}

#[wasm_bindgen]
pub fn aes_decrypt(content: String, key: String) -> String {
    let pass = utils::crypto_utils::keccak256_hash(&key);
    let msg_hex = utils::crypto_utils::aes_decrypt(&content, &pass);
    msg_hex
}

#[wasm_bindgen]
pub fn rencrypt(pk: String, msg: String) -> String {
    let public_key = match PublicKey::from_str(&msg) {
        Ok(v) => v,
        Err(e) => panic!("error parse publickey: {}", e),
    };
    let msg_encrypt = match zencrypt(public_key, &msg) {
        Ok(v) => v,
        Err(err) => panic!("error encrypt: {:?}", err),
    };
    msg_encrypt
}

#[wasm_bindgen]
pub fn rdecrypt(sk: String, msg: String) -> String {
    let private_key = match SecretKey::from_str(&sk) {
        Ok(v) => v,
        Err(e) => panic!("error parse publickey: {}", e),
    };
    let msg_decrypt = match zdecrypt(private_key, &msg) {
        Ok(v) => v,
        Err(err) => panic!("error encrypt: {:?}", err),
    };
    msg_decrypt
}

#[wasm_bindgen]
pub fn generate_qr(content: String) -> String {
    QR::parse(&content)
}

#[wasm_bindgen]
pub fn hex_deflate(content: String) -> String {
    hex_to_base64(&content)
}

#[wasm_bindgen]
pub fn hex_inflate(content: String) -> String {
    base64_to_hex(&content)
}

#[wasm_bindgen]
pub fn password_hash(pass: String, salt: String) -> String {
    hash_password(&pass, &salt)
}

#[wasm_bindgen]
pub fn hash_pass_svr(content: String) -> String {
    let msg_hex = utils::crypto_utils::hash_pass_svr(&content);
    msg_hex
}

#[wasm_bindgen]
pub fn password_verify(pass: String, pass_hash: String) -> bool {
    verify_password(&pass, &pass_hash)
}
