use aes_gcm::{
    aead::{generic_array::GenericArray, Aead, KeyInit},
    Aes256Gcm,
};
use anyhow::Result;
use argon2::{
    password_hash::{rand_core::OsRng, PasswordHash, PasswordHasher, PasswordVerifier, SaltString},
    Argon2,
};
use base64::{engine::general_purpose, Engine as _};
use ecies::{decrypt, encrypt};
use rand::prelude::*;
use scrypt::{scrypt, ScryptParams};
use secp256k1::{PublicKey, SecretKey};
use std::str;
use tiny_keccak::keccak256;

pub fn random_argon2_salt() -> String {
    let salt: SaltString = SaltString::generate(&mut OsRng);
    salt.as_str().to_string()
}
pub fn hash_password(pass: &str, salt: &str) -> String {
    // let salt: SaltString = SaltString::generate(&mut OsRng);
    let salt = SaltString::from_b64(salt).expect("error parse salt");
    let password = pass.as_bytes();
    let password_hash: String = match Argon2::default().hash_password(password, &salt) {
        Ok(v) => v.to_string(),
        Err(e) => panic!("error hash password: {}", e),
    };
    general_purpose::STANDARD_NO_PAD.encode(&password_hash)
}

pub fn scrypt_hash(pass: &str, salt: &str, n: u8, r: u32, p: u32, size: u32) -> String {
    let password = pass.as_bytes();
    let salt = hex::decode(salt).expect("error decoding hex string");
    let params = ScryptParams::new(n, r, p).expect("error create scrypt params");
    let mut output = vec![0u8; size as usize];
    scrypt(password, &salt, &params, &mut output).expect("error scrypt");
    general_purpose::STANDARD_NO_PAD.encode(output)
}

pub fn verify_password(pass: &str, password_hash: &str) -> bool {
    let str_tmp = match general_purpose::STANDARD_NO_PAD.decode(password_hash) {
        Ok(v) => v,
        Err(e) => panic!("error decode base64 str: {}", e),
    };
    let s = match str::from_utf8(&str_tmp) {
        Ok(v) => v,
        Err(e) => panic!("Invalid UTF-8 sequence: {}", e),
    };
    let parsed_hash = match PasswordHash::new(&s) {
        Ok(v) => v,
        Err(e) => panic!("error parse password hash: {}", e),
    };

    let password = pass.as_bytes();
    Argon2::default()
        .verify_password(password, &parsed_hash)
        .is_ok()
}

pub fn keccak256_hash(str: &str) -> String {
    let data = str.as_bytes();
    let hasher = keccak256(data);
    hex::encode(&hasher)
}

pub fn hash_pass_client(str: &str) -> String {
    let message = str.as_bytes();
    let mut eth_message = format!("\x22cebg password pre hash:\n{}", message.len()).into_bytes();
    eth_message.extend_from_slice(message);

    let hasher = keccak256(&eth_message);
    hex::encode(&hasher)
}

pub fn hash_pass_svr(str: &str) -> String {
    let message = str.as_bytes();
    let mut eth_message = format!("\x23cebg email regist:\n{}", message.len()).into_bytes();
    eth_message.extend_from_slice(message);

    let hasher = keccak256(&eth_message);
    hex::encode(&hasher)
}
//
// This function is used to encrypt a string using AES-256-GCM algorithm
pub fn aes_encrypt(str: &str, key: &str) -> String {
    // Convert the input string to bytes
    let str_data = str.as_bytes();

    // Decode the input key from hex to bytes and create a GenericArray object
    let key = hex::decode(&key).expect("Invalid hex string");
    let key = GenericArray::from_slice(&key);

    // Create an AES-256-GCM cipher with the specified key
    let cipher = Aes256Gcm::new(&key);

    // Generate a random nonce of 96 bits (12 bytes)
    let mut nonce_data = [0u8; 12];
    thread_rng().fill_bytes(&mut nonce_data);
    let nonce = GenericArray::from_slice(&nonce_data);

    // Encrypt the data using the cipher and nonce
    let ciphertext = match cipher.encrypt(nonce, str_data) {
        Ok(v) => v,
        Err(e) => panic!("error encrypt: {}", e),
    };
    // Encode the ciphertext and nonce as base64 strings
    let cipher_str = general_purpose::STANDARD_NO_PAD.encode(&ciphertext);
    let nonce_str = general_purpose::STANDARD_NO_PAD.encode(&nonce_data);
    // Concatenate the encoded ciphertext and nonce with a period separator and return the result
    nonce_str + cipher_str.as_str()
}

pub fn aes_decrypt(str: &str, key: &str) -> String {
    let str_data = match general_purpose::STANDARD_NO_PAD.decode(str) {
        Ok(v) => v,
        Err(e) => panic!("error decode base64 str: {}", e),
    };
    if str_data.len() < 12 {
        panic!("error decrypt: invalid data");
    }
    let v = 12;
    let nonce_data = &str_data[..v];
    let cipher_data = &str_data[v..];

    let nonce = GenericArray::from_slice(&nonce_data);
    let key = hex::decode(&key).expect("Invalid hex string");
    let key = GenericArray::from_slice(&key);
    let cipher = Aes256Gcm::new(&key);

    let plaintext = match cipher.decrypt(nonce, cipher_data) {
        Ok(v) => v,
        Err(e) => panic!("error decrypt: {}", e),
    };
    let plaintext = str::from_utf8(&plaintext).expect("err convert to utf8");
    plaintext.to_string()
}
// encrypt content with wallet public key
pub fn zencrypt(pk: PublicKey, msg: &str) -> Result<String> {
    let pk = &pk.serialize();
    let msg = msg.as_bytes();
    // println!("msg before encrypt: {:?}", msg);
    let msg_encrypt = match encrypt(pk, &msg) {
        Ok(v) => v,
        Err(e) => panic!("error encrypt content: {}", e),
    };
    // println!("msg after encrypt: {:?}", &msg_encrypt);
    let str_encrypt = hex::encode(&msg_encrypt);
    Ok(str_encrypt)
}
// decrypt content with wallet private key
pub fn zdecrypt(sk: SecretKey, msg1: &str) -> Result<String> {
    let sk = sk.secret_bytes();
    let mut msg: String = msg1.clone().to_string();
    if msg.len() % 2 == 1 {
        msg = "0".to_owned() + &msg;
    }
    println!("msg to decrypt: {:?}", &msg);
    let msg = match hex::decode(&msg) {
        Ok(v) => v,
        Err(e) => panic!("error decode hex str: {}", e),
    };
    let msg_decrypt = match decrypt(&sk, &msg) {
        Ok(v) => v,
        Err(e) => panic!("error decrypt content: {}", e),
    };
    // println!("msg after decrypt: {:?}", &msg_decrypt);
    // let msg_decrypt = hex::encode(msg_decrypt);
    let str_decrypt = match str::from_utf8(&msg_decrypt) {
        Ok(v) => v,
        Err(e) => panic!("Invalid UTF-8 sequence: {}", e),
    };
    let result = str_decrypt.to_string();

    Ok(result)
}
